# xorn.geda.netlist - gEDA Netlist Extraction and Generation
# Copyright (C) 1998-2010 Ales Hvezda
# Copyright (C) 1998-2010 gEDA Contributors (see ChangeLog for details)
# Copyright (C) 2013-2015 Roland Lutz
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.

## \namespace xorn.geda.netlist.pp_bus
## Post-processing: Bus pins and netnames.

from gettext import gettext as _
import xorn.geda.netlist.blueprint

def parse_pinrange(s, log):
    if s[-1] == ']':
        try:
            pos = s.index('[')
        except ValueError:
            log.error(_("invalid pin range format: %s") % s)
            return None
        else:
            stem, s = s[:pos], s[pos + 1:-1]
    else:
        stem = None

    l = []

    for s1 in s.split(','):
        pos = s1.find('..')
        if pos != -1:
            try:
                start = int(s1[:pos])
                stop = int(s1[pos + 2:])
            except ValueError:
                log.error(_("non-numeric pin range: %s") % s1)
                return None

            if start > stop:
                l += map(str, reversed(xrange(stop, start + 1)))
            else:
                l += map(str, xrange(start, stop + 1))
            continue

        pos = s1.find(':')
        if pos != -1:
            try:
                start = int(s1[:pos])
                stop = int(s1[pos + 1:])
            except ValueError:
                log.error(_("non-numeric pin range: %s") % s1)
                return None

            if start > stop:
                l += map(str, reversed(xrange(stop, start)))
            else:
                l += map(str, xrange(start, stop))
            continue

        if s1.startswith(' ') or s1.endswith(' '):
            log.error(_("leading/trailing spaces in pin name: \"%s\"") % s1)
            return None

        l.append(s1)

    if len(set(l)) != len(l):
        log.error(_("overlapping pin range: %s") % s)
        return None

    if stem is None:
        return l
    return [stem + l for l in l]

def postproc_blueprints(netlist):
    for schematic in netlist.schematics:
        for component in schematic.components:
            portname = component.get_attribute('portname', None)

            for pin in component.pins[:]:
                if pin.ob is None or not pin.ob.data().is_bus:
                    continue

                # found a bus pin

                if portname is not None:
                    # remove original bus pin
                    schematic.port_pins[portname].remove(pin)
                    # on a port symbol, use its portname= attribute
                    pinrange = portname
                elif component.composite_sources:
                    # on a subsheet symbol, use the pinlabel=
                    pinrange = pin.get_attribute('pinlabel', None)
                    if pinrange is None:
                        # pp_hierarchy will throw an error later
                        continue
                else:
                    # otherwise, use the pinnumber
                    try:
                        pinrange = pin.get_attribute('pinnumber')
                    except KeyError:
                        pin.error(_("pinnumber missing on bus pin"))
                        continue

                pinnumbers = parse_pinrange(pinrange, pin)
                if pinnumbers is None:
                    continue

                assert pin.bus_bits is None
                pin.bus_bits = []

                for pinnumber in pinnumbers:
                    try:
                        bit_pin = component.pins_by_number[pinnumber]
                    except KeyError:
                        bit_pin = xorn.geda.netlist.blueprint.Pin(
                            component, None)
                        if portname is not None:
                            try:
                                port_pins = schematic.port_pins[pinnumber]
                            except KeyError:
                                port_pins = schematic.port_pins[pinnumber] = []
                            port_pins.append(bit_pin)
                        elif component.composite_sources:
                            bit_pin.label = pinnumber
                        else:
                            bit_pin.number = pinnumber
                            component.pins_by_number[pinnumber] = bit_pin
                    else:
                        pin.error(_("duplicate pinnumber \"%s\"") % pinnumber)

                    pin.bus_bits.append(bit_pin)

                # remove the bus pin
                component.pins.remove(pin)
                assert component.pins_by_ob[pin.ob] == pin
                del component.pins_by_ob[pin.ob]
                # ignore pins_by_pinseq and pins_by_number
                # because bus pins aren't slotted (yet)

    for schematic in netlist.schematics:
        for net in schematic.nets:
            if not net.is_bus:
                continue

            parsed_names = [parse_pinrange(name, net) for name in net.names]

            # determine bus width
            bus_widths = []
            for pin in net.pins:
                if pin.bus_bits is not None \
                        and len(pin.bus_bits) not in bus_widths:
                    bus_widths.append(len(pin.bus_bits))
            for parsed_name in parsed_names:
                if parsed_name is not None \
                        and len(parsed_name) not in bus_widths:
                    bus_widths.append(len(parsed_name))
            if not bus_widths:
                # purely decorative bus
                continue
            if len(bus_widths) > 1:
                net.error(_("bus width mismatch: %s") %
                          _(" vs. ").join(str(width) for width in bus_widths))
                continue
            bus_width, = bus_widths

            net.bus_bits = [xorn.geda.netlist.blueprint.Net(schematic, [])
                            for i in xrange(bus_width)]
            for bus_bit, names in zip(net.bus_bits, zip(*parsed_names)):
                bus_bit.names += names

            for pin in net.pins:
                if pin.bus_bits is None:
                    continue

                for net_bit, pin_bit in zip(net.bus_bits, pin.bus_bits):
                    if pin_bit.net is not None:
                        assert pin_bit.net == net_bit
                        continue
                    pin_bit.net = net_bit
                    net_bit.pins.append(pin_bit)

        schematic.nets = [net for net in schematic.nets if not net.is_bus]

    for schematic in netlist.schematics:
        for component in schematic.components:
            for pin in component.pins:
                if pin.ob is None or not pin.ob.data().is_bus:
                    continue
                if pin.bus_bits is None:
                    continue

                for pin_bit in pin.bus_bits:
                    if pin_bit.net is None:
                        net = xorn.geda.netlist.blueprint.Net(schematic, [])
                        pin_bit.net = net
                        net.pins.append(pin_bit)
