# xorn.geda.netlist - gEDA Netlist Extraction and Generation
# Copyright (C) 1998-2010 Ales Hvezda
# Copyright (C) 1998-2010 gEDA Contributors (see ChangeLog for details)
# Copyright (C) 2013-2015 Roland Lutz
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.

## \namespace xorn.geda.netlist.pp_hierarchy
## Post-processing: Hierarchy traversal.

from gettext import gettext as _
import xorn.geda.attrib

def postproc_blueprints(netlist):
    for schematic in netlist.schematics:
        schematic.port_pins = {}

        for component in schematic.components:
            portname = component.get_attribute('portname', None)
            if portname is None:
                continue

            if component.get_attribute('refdes', None) is not None:
                component.error(_("refdes= and portname= attributes "
                                  "are mutually exclusive"))
            if component.get_attribute('netname', None) is not None:
                component.error(_("netname= and portname= attributes "
                                  "are mutually exclusive"))
            if xorn.geda.attrib.search_all(component.ob, 'net'):
                component.error(_("portname= and net= attributes "
                                  "are mutually exclusive"))

            if not component.pins:
                component.error(_("I/O symbol doesn't have pins"))
            if len(component.pins) > 1:
                component.error(_("multiple pins on I/O symbol"))
            for pin in component.pins:
                if pin.number is not None or pin.ob.attached_objects():
                    pin.error(_("pin on I/O symbol can't have attributes"))

            if component.is_graphical:
                component.error(_("I/O symbol can't be graphical"))

            try:
                port_pins = schematic.port_pins[portname]
            except KeyError:
                port_pins = schematic.port_pins[portname] = []
            port_pins += component.pins

            component.has_portname_attrib = True

## Connect subsheet I/O ports to the instantiating component's pins.
#
# Disconnect all connections from and to composite components as
# they have been replaced with the actual subschematics.
#
# remove all composite components and ports

def postproc_instances(netlist):
    remove_components = set()

    for component in netlist.components:
        if not component.blueprint.composite_sources:
            continue

        # collect potential old-style ports
        refdes_dict = {}
        for subsheet in component.subsheets:
            for potential_port in subsheet.components:
                try:
                    l = refdes_dict[potential_port.blueprint.refdes]
                except KeyError:
                    l = refdes_dict[potential_port.blueprint.refdes] = []
                l.append(potential_port)

        for cpin in component.cpins:
            label = cpin.blueprint.label
            if label is None:
                cpin.error(_("pin on composite component is missing a label"))
                continue

            dest_net = cpin.local_net.net

            # search for the matching port
            port_pins = [
                subsheet.components_by_blueprint[port_pin.component]
                        .cpins_by_blueprint[port_pin]
                for subsheet in component.subsheets
                for port_pin in subsheet.blueprint.port_pins.get(label, [])]

            for port in refdes_dict.get(label, []):
                # found an old-style port
                if not port.cpins:
                    port.error(_("I/O symbol doesn't have pins"))
                    continue
                if len(port.cpins) > 1:
                    port.error(_("multiple pins on I/O symbol"))
                    continue
                if port.blueprint.is_graphical:
                    port.error(_("I/O symbol can't be graphical"))

                if port.blueprint.has_netname_attrib:
                    port.error(_("netname= attribute can't be used "
                                 "on an I/O symbol"))
                if xorn.geda.attrib.search_all(port.blueprint.ob, 'net'):
                    port.error(_("net= attribute can't be used "
                                 "on an I/O symbol"))
                port_pins += port.cpins

            if not port_pins:
                cpin.warn(_("missing I/O symbol for port `%s' "
                            "inside schematic") % label)
            elif len(port_pins) > 1:
                cpin.warn(_("multiple I/O symbols for port `%s' "
                            "inside schematic") % label)

            for port_pin in port_pins:
                src_net = port_pin.local_net.net

                # merge nets
                if src_net != dest_net:
                    src_net.merge_into(dest_net)
                    dest_net.component_pins += src_net.component_pins
                    del src_net.component_pins[:]

                # remove port component
                remove_components.add(port_pin.component)

                port_pin.local_net.cpins.remove(port_pin)
                dest_net.component_pins.remove(port_pin)

            # After the pin has been connected, remove it.
            cpin.local_net.cpins.remove(cpin)
            dest_net.component_pins.remove(cpin)

        # After all pins have been connected, remove the component.
        remove_components.add(component)

    netlist.components = [component for component in netlist.components
                          if component not in remove_components]
    for component in remove_components:
        component.sheet.components.remove(component)
        del component.sheet.components_by_blueprint[component.blueprint]

    for component in netlist.components:
        if component.blueprint.has_portname_attrib:
            component.error(_("unmatched I/O symbol"))
